/*! @file
 * @brief
 * implementation of concepts for client-sided I/O to the server
 */

#include "Client_IO.H"

// -------------------------------------------------------------------------

string CClientPolynomFetcher::buffer;

void * CClientPolynomFetcher::THREAD_fetch_polynom(void *)
{
  static unsigned int PolyInterval = 1000;
  static time_t prev_time = 0;
  if (PolyInterval>500 && time(NULL)>prev_time+300) PolyInterval=100*(PolyInterval/200);
  else if (time(NULL)<prev_time+120) PolyInterval+=100*(PolyInterval/200);
  prev_time=time(NULL);

  unix_io_stream communication_stream2(communication_name, server_port);
  cout_network << "Fetching mpqs polynomial interval (" << PolyInterval << ")..." << endl;
  communication_stream2 << "Polynom? " << PolyInterval << endl;
  buffer.clear();
  while (!communication_stream2.eof())
   {
     char c;
     communication_stream2.get(c);
     buffer+=c;
   }
  cout_network << "background thread: polynomial fetched." << endl;
  return NULL;
}

void CClientPolynomFetcher::fetch(mpz_t UpperBound_D)
{
  static pthread_t thread_fetch_polynom;
  int retcode;

  static bool first_time_first_love = true;
  if (first_time_first_love)
   {
     first_time_first_love=false;
     retcode = pthread_create(&thread_fetch_polynom, NULL, THREAD_fetch_polynom, NULL);
     if (retcode != 0)
      {
        cerr << "CClientPolynomFetcher: pthread_create failed!" << endl;
        exit(1);
      }
   }

  // Thread joinen
  cout_network << "Polynomfetch join..." << endl;
  retcode = pthread_join(thread_fetch_polynom, NULL);
  if (retcode != 0)
   {
     cerr << "CClientPolynomFetcher: Joining thread failed!" << endl;
     exit(1);
   }

  cout << "Setting mpqs polynomial interval..." << endl;

  // read in polynomial data from internal string buffer
  istringstream is(buffer);
  Polynom.load(is);
  cout << "Old UpperBound_D = " << UpperBound_D << endl;
  is >> UpperBound_D;
  cout << "New UpperBound_D = " << UpperBound_D << endl;

  // and fetch the next polynomial (threaded in background)
  retcode = pthread_create(&thread_fetch_polynom, NULL, THREAD_fetch_polynom, NULL);
  if (retcode != 0)
   {
     cerr << "CClientPolynomFetcher::pthread_create failed!" << endl;
     exit(1);
   }
}


// ----------------------------------------------------------------------


CMutex CClientDynamicFactorFetcher::Mutex;
queue<int> CClientDynamicFactorFetcher::buffer;

void * CClientDynamicFactorFetcher::THREAD_fetch_DynamicFactors(void *)
{
 int sleep_secs = 100;
 while(true)
  {
   {
     unix_io_stream communication_stream2(communication_name, server_port);
     cout << "Fetching dynamic factors from server..." << endl;
     static unsigned int dynfac_pos = 0;
     communication_stream2 << "DynamicFactors?_ab_index " << dynfac_pos << endl;

     int factor = 0;
     int counter = 0;
     while (true) // fetching dynamic relations
      {
        //communication_stream2 >> factor;
        // cout << factor << "\r" << flush;
        char c[4];
        if (communication_stream2.peek()==EOF)
         {
           // magical mystery tour...
           // it seems that EOF is set sometimes although the stream
           // is still fully readable...
           // for that reason we clear the stream just before the
           // real read...
           if ( Cpoll(communication_stream2).readable_chars_within(128,2000) < 4)
             cerr << "possibly failing..." << endl;
           communication_stream2.clear();
         }
        communication_stream2.read(c,4);
        factor=static_cast<unsigned int>(static_cast<unsigned char>(c[0]));
        factor|=static_cast<unsigned int>(static_cast<unsigned char>(c[1]))<<8;
        factor|=static_cast<unsigned int>(static_cast<unsigned char>(c[2]))<<16;
        factor|=static_cast<unsigned int>(static_cast<unsigned char>(c[3]))<<24;
        if (factor<=0 || communication_stream2.fail())
         {
           MARK;
           break;
         }
        Mutex.lock(); buffer.push(factor); Mutex.unlock();
        ++counter;
      }
     dynfac_pos+=counter;
     cout_network << "background thread: " << counter << " dynamic factors fetched." << endl;
     if (counter>200) sleep_secs-=sleep_secs/4;
     else if (counter<20 && sleep_secs<10000) sleep_secs+=(sleep_secs+3)/4;
   }
   cout_network << "next dynamic factor fetch request in " << sleep_secs << " seconds." << endl;
   sleep(sleep_secs); // wait a few seconds...
  }
 return NULL;
}


void CClientDynamicFactorFetcher::fetch()
{
  static pthread_t thread_fetch_DynamicFactors;

  int retcode;
  static bool first_time_first_love = true;

  if (first_time_first_love)
   {
     first_time_first_love=false;
     retcode = pthread_create(&thread_fetch_DynamicFactors, NULL, THREAD_fetch_DynamicFactors, NULL);
     if (retcode != 0)
      {
        cerr << "CClientDynamicFactorFetcher: pthread_create failed!" << endl;
        exit(1);
      }
   }

  Mutex.lock();
  if (buffer.empty())
   {
     Mutex.unlock();
     return;
   }

#ifdef VERBOSE
  cout << "Inserting " << buffer.size() << " dynamic factors..." << endl;
#endif

  // read in dynamic faktors from internal class buffer
  TDynamicFactorRelation relation;
  relation.fpos=0; relation.factor=0;
  while (!buffer.empty()) // get dynamic relation
   {
     relation.factor=buffer.front(); buffer.pop();
     // cout << relation.factor << "\r" << flush;
     relation.SQRT_kN=SQRT_kN_mod_PrimeNumber(relation.factor);
     DynamicFactorRelations.insert(relation);
   }
  Mutex.unlock();
}


// ---------------------------------------------------------------------------

volatile bool CClientRelation_Delivery::Sending_Relations = true; // consumer/producer trigger
int  CClientRelation_Delivery::Sende_Nr = 0;
string CClientRelation_Delivery::file_name;
ifstream CClientRelation_Delivery::ConsumerFile;
CMutex CClientRelation_Delivery::Consumer;


const string CClientRelation_Delivery::create_unique_tempfilename(const string servername)
{
  ostringstream ostr;
  pid_t pid = getpid();
  struct utsname buf;
  uname(&buf);
  ostr << TempDir << "/qsieve-" << buf.nodename
       << "-" << servername << "_" << pid;
  return ostr.str();
}

void CClientRelation_Delivery::init(void)
{
  // first check whether the host is reachable
  // and get the hostname

#ifdef CYGWIN_COMPAT
  struct hostent *hp;
  hp = gethostbyname(communication_name.c_str());
  if (hp == NULL)
    {
      cerr << "Unknown host " << communication_name << endl;
      exit (1);
    }
#else
  // refer "man getaddrinfo" and "man socket" 
  struct addrinfo hints; // our wishes are placed here
  memset(&hints,0,sizeof(hints)); // must be zeroed for default options
  hints.ai_family=PF_INET; // we want IPv4 as protocol
  hints.ai_socktype=SOCK_STREAM; // and we need a stream, not datagram!
  struct addrinfo *addrinfo_res = NULL; // here the result will be delivered
  const int retval = getaddrinfo(communication_name.c_str(),NULL,&hints,&addrinfo_res);
  if ( retval || addrinfo_res==NULL ) // any error?
   {
     cerr << "can't reach " << "\"" <<  communication_name << "\"" << endl;
     cerr << "Error given by getaddrinfo: " << endl;
     cerr << gai_strerror(retval) << endl;
     exit(1);
   }
  if (addrinfo_res->ai_socktype!=SOCK_STREAM) // we got a "stream"-protocol?
   {
     cerr << "provided protocol doesn't support SOCK_STREAM" << endl;
     exit(1);
   }

  freeaddrinfo(addrinfo_res); // free ressources allocated by getaddrinfo
#endif

  // create an unique tempfilename based on the hostname
  file_name = create_unique_tempfilename(communication_name);
}

void CClientRelation_Delivery::cleanup_files(void)
{
  string name;
  name = file_name+"A";
  unlink(name.c_str());
  name = file_name+"B";
  unlink(name.c_str());
}

void * CClientRelation_Delivery::THREAD_transmit_Relations(void *)
{
  // important: this thread shall not run concurrently to itself!
  static int retval; // &retval is returned to thread who is joining
  string s;
  char line[1024];

  retval = -1; // default: return an error
  Consumer.lock();

  unix_io_stream communication_stream2(communication_name, server_port);
  
  communication_stream2 << "NewRelations! " << kN << endl;
  // verify, that relations belong to our factorization
  communication_stream2 >> s;
  if (s!="proceed")
   {
     cerr << "Oops! Server will not accept my relations: " << endl;
     cerr << s << endl;
     goto done;
   }

  if (!ConsumerFile.is_open())
   {
     cerr << "ConsumerFile is not open!" << endl;
     goto done;
   }

transmit:
  while (ConsumerFile.peek()!=EOF)
   {
     int count=1000+1; // send maximal count-1 relationen in per block
     const int max_Uebertragungsdauer = 8; // Maximal duration of a transmission for a block in seconds (to prevent life-locks...)
     const time_t Uebertragungsstart = time(NULL);
     while (ConsumerFile.peek()!=EOF && --count && time(NULL)-Uebertragungsstart<=max_Uebertragungsdauer)
      {
        communication_stream2 << "Relation! ";
	while (ConsumerFile.peek()!='\n')
	 {
	   ConsumerFile.get(line,sizeof(line),'\n');
	   communication_stream2 << line;
	 }
	char c; ConsumerFile.get(c);
	communication_stream2 << endl;
      }
     
     if (communication_stream2.fail())
      {
	MARK;
        cerr << "stream state indicates that transmission had failed!" << endl;
        goto done;
      }

     if (ConsumerFile.peek()!=EOF)
      {
        // synchronize communication with server
        communication_stream2 << "Relationsblock_Sync " << flush;
        communication_stream2 >> s;
        if (s!="synced.")
         {
           cerr << "sync with server failed: " << endl; cerr << s << endl;
           goto done;
         }
      }
   }
  ConsumerFile.clear(); ConsumerFile.close();

  // synchronize communication with server
  communication_stream2 << "Relationsblock_Sync " << flush;
  communication_stream2 >> s;
  if (s!="synced.")
   {
     cerr << "sync with server failed: " << endl; cerr << s << endl;
     goto done;
   }

  Consumer.unlock();
  sleep(5);
  Consumer.lock();
  if (ConsumerFile.is_open()) goto transmit;

  communication_stream2 << "Relationsblock_Ende " << flush;
  communication_stream2 >> s;
  if (s!="empfangen.")
   {
     cerr << "Transmission of relations not acknowledged!" << endl;
     goto done;
   }
  if (s=="ignoriert!")
   {
     cerr << "Oops! - Transmitted relations were rejected by the server!" << endl;
     goto done;
   }
  sleep(60); // prevent too many connections within a few seconds from fast clients
  retval=0; // clear error state
done:
  Sending_Relations=true; // bump! a new thread is allowed now...
  Consumer.unlock();
  return &retval;
}

void CClientRelation_Delivery::go(void)
{
  if (!communication_stream)
   {
     MARK;
     cerr << "communication stream is in error state! aborting..." << endl;
     exit(1);
   }

#ifdef STL_STREAM_workaround
  if ( (communication_stream.tellp()==std::streampos(0)) || (communication_stream.tellp()==std::streampos(-1)) )
#else
  if (communication_stream.tellp()<1)
#endif
   {
#ifdef VERBOSE
     MARK;
     cout << "no produced data..." << endl;
#endif
     return;
   }

  static pthread_t thread_Relationen_versenden;
  static const char* const  SendeSuffix[2] = { "A", "B" };

  if (Sending_Relations)
   {
     // ready to transmit relations to server:
     // - set flag to trigger transmission: -> Sending_Relations=false
     // - create thread for consuming and transmitting the data
     // - the consumer-thread will reset the flag after he has finished his job
     // - if the thread indicates, he had done his work, join him and create a new one
     
     Sending_Relations = false;
     
     // close file (producer)
     communication_stream.close();

     // open file (consumer)
     const string ConsumerFileName = file_name + SendeSuffix[Sende_Nr];
     ConsumerFile.open(ConsumerFileName.c_str(), ios::in);
     if (!ConsumerFile)
      {
        MARK; cerr << "opening consumer file \"" << ConsumerFileName << "\" failed!" << endl;
        exit(1);
      }

     int retcode;
     void* thread_retval = NULL;

     static bool first_time_first_love = true;

     if (first_time_first_love)
      { // vor dem ersten Thread nicht joinen, ...
        first_time_first_love=false;
        retcode = pthread_create(&thread_Relationen_versenden, NULL,
                    CClientRelation_Delivery::THREAD_transmit_Relations, NULL);
      }
     else
      { // ... danach schon...
        retcode = pthread_join(thread_Relationen_versenden, &thread_retval);
        if (retcode != 0)
         {
           cerr << "Joining Consumer-Thread failed!" << endl;
           exit(1);
         }
        else
         {
           if (thread_retval && *(static_cast<int*>(thread_retval)) ) // was a value returned?
            {
              // this means, an error occurred...
              MARK;
              cerr << "Consumer-Thread returned an error value: " << *(static_cast<int*>(thread_retval)) << endl;
              exit(1);
            }
         }
        retcode = pthread_create(&thread_Relationen_versenden, NULL,
                    CClientRelation_Delivery::THREAD_transmit_Relations, NULL);
      }

     if (retcode != 0)
      {
        cerr << "Thread for transmitting relations failed!" << endl;
        exit(1);
      }

     Sende_Nr = 1 - Sende_Nr;
     // open new file (producer)
     const string name = file_name + SendeSuffix[Sende_Nr];
     communication_stream.open(name.c_str() ,ios::out|ios::trunc);
     if (!communication_stream)
      {
        MARK; cerr << "opening communication_stream \"" << name << "\" failed!" << endl;
        exit(1);
      }
   }
  else
   {
     // consumer thread is transmitting or waiting for data or just exiting
     if (Consumer.trylock())
      {
        // remember, that it is unlikely, but possible, that Sending_Relations==true now,
        // (we could get interrupted after checking its value!)
 
        if (Sending_Relations || ConsumerFile.is_open()) { Consumer.unlock(); return; }

        // well, now we know that he's waiting :-)
        // and that means that we can give him relations to consume...

        // close file (producer)
        communication_stream.close();

        // open file (consumer)
        const string ConsumerFileName = file_name + SendeSuffix[Sende_Nr];
        ConsumerFile.open(ConsumerFileName.c_str(), ios::in);
        if (!ConsumerFile)
         {
           MARK; cerr << "opening consumer file \"" << ConsumerFileName << "\" failed!" << endl;
           exit(1);
         }

        Sende_Nr = 1 - Sende_Nr;
        // open new file (producer)
        const string name = file_name + SendeSuffix[Sende_Nr];
        communication_stream.open(name.c_str() ,ios::out|ios::trunc);
        if (!communication_stream)
         {
           MARK; cerr << "opening communication_stream \"" << name << "\" failed!" << endl;
           exit(1);
         }

        Consumer.unlock();
      }
   }
}
