/*! @file
 * @brief
 * configfile related functions
 */

void Read_ConfigFile(void)
{
  ifstream in(ConfigFile.c_str());
  char InputLine[200];
  string zeile;
  if (in)
    {
#ifdef VERBOSE_INFO
      cout << "Reading configuration file" << endl;
#ifdef IS_CLIENT
      cout << "Some server related options may have no effect to clients." << endl;
#endif
#endif
      while (!in.eof())
	{
	  in.getline(InputLine,sizeof(InputLine),'\n');
	  if (InputLine[0]=='#') continue; // read over comment lines
	  zeile=InputLine;
          while (zeile[0]==' ') zeile.erase(0,1);
          if (zeile.empty() || zeile[0]=='#') continue; // read over empty lines and comment lines

          // process keywords
          istringstream is(zeile);
          string Keyword;
          is >> Keyword;
          if (Keyword=="DynamicRelationsFile")
           {
             string s;
             is >> s;
             if (s!="=")
              {
                cerr << "Error in " << ConfigFile  << "!" << endl;
                exit(1);
              }
             is >> DynamicRelationsFile;
             continue;
           }
          if (Keyword=="WorkDir")
           {
             string s;
             is >> s;
             if (s!="=")
              {
                cerr << "Error in " << ConfigFile  << "!" << endl;
                exit(1);
              }
             is >> s;
#ifdef VERBOSE_NOTICE
             cout << "Setting current working directory to \"" << s << "\"." << endl;
#endif
	     if (chdir(s.c_str())!=0)
              {
                cerr << "Changing working directory failed!" << endl;
                exit(2);
              };
             continue;
           }
          if (Keyword=="TempDir")
           {
             string s;
             is >> s;
             if (s!="=")
              {
                cerr << "Error in " << ConfigFile  << "!" << endl;
                exit(1);
              }
             is >> s;
#ifdef VERBOSE_NOTICE
             cout << "Setting TempDir to \"" << s << "\"." << endl;
#endif
             TempDir=s; 
             continue;
           }
          if (Keyword=="XMLStatusFile")
           {
             string s;
             is >> s;
             if (s!="=")
              {
                cerr << "Error in " << ConfigFile  << "!" << endl;
                exit(1);
              }
             is >> s;
#ifdef VERBOSE_NOTICE
             cout << "Setting current XMLStatusFile to \"" << s << "\"." << endl;
#endif
             XML_StatusFile=s; 
             continue;
           }
          if (Keyword=="SkipFermat")
           {
	     SkipFermat=true;
             continue;
           }
          if (Keyword=="SkipPhi")
           {
	     SkipPhi=true;
             continue;
           }
          if (Keyword=="SkipFibonacci")
           {
	     SkipFibonacci=true;
             continue;
           }
          if (Keyword=="SkipEasyECM")
           {
	     SkipEasyECM=true;
             continue;
           }
          if (Keyword=="SkipECM")
           {
	     SkipECM=true;
             continue;
           }

        }
    }
  else
    {
       cout << ConfigFile << " not found! Using default values!" << endl;
#ifndef IS_CLIENT
       static char ch = 'N';
       if (ch!='Y')
         {
           cout << "Bitte mit [Y]es besttigen!" << endl;
           cout << "Please confirm with [Y]es." << endl;
           cin >> ch;
           if (ch!='Y') exit(1);
         }
#endif
    }

#ifdef VERBOSE_INFO
  char cwd[4096];
  cout << "Current working directory: " << getcwd(cwd,4096) << endl;
  cout << "Configuration file: " << ConfigFile << endl;
  cout << "temp directory: " << TempDir << endl;
  cout << "DynamicRelationsFile: " << DynamicRelationsFile << endl;
#endif
#ifndef IS_CLIENT
  // clients do report factors to the server and not to this file,
  // so only stand-alone versions and servers need this:
  if (!Factorization_to_file.is_open()) Factorization_to_file.open(FactorizationsFile.c_str(),ios::out|ios::app);
#endif
}


#ifndef IS_CLIENT
// clients will be configured by the server, so only stand-alone and server versions need this...
void tune_parameters(const unsigned int DecimalDigits)
{
  ifstream in(ConfigFile.c_str());
  char InputLine[200];
  string zeile;
  unsigned int Stellen, gewaehlte_Stellenzahl=0;
  if (in)
    {
#ifdef VERBOSE_INFO
      cout << "Reading configuration file" << endl; 
#endif
      while (!in.eof())
	{
	  in.getline(InputLine,sizeof(InputLine),'\n');
	  if (InputLine[0]=='#') continue; // read over comment lines
	  zeile=InputLine;
          while (zeile[0]==' ') zeile.erase(0,1);
          if (zeile.empty() || zeile[0]=='#') continue; // read over empty lines / comment lines

          // process keywords
          istringstream is(zeile);
          string Keyword;
          is >> Keyword;
	  if (Keyword=="SkipFermat") continue;
	  if (Keyword=="SkipPhi") continue;
	  if (Keyword=="SkipFibonacci") continue;
          if (Keyword=="SkipEasyECM") continue;
	  if (Keyword=="SkipECM") continue;
          if (Keyword=="DynamicRelationsFile" || Keyword=="WorkDir" || Keyword=="TempDir" || Keyword=="XMLStatusFile")
           {
             string s;
             is >> s;
             if (s!="=")
              {
                cerr << "Error in " << ConfigFile  << "!" << endl;
                exit(1);
              }
             // is >> DynamicRelationsFile; // do not read anything here!
             continue;
           }

          // otherwise: line contains parameter for tuning
	  Stellen = atoi(zeile.substr(0, zeile.find(":")).c_str());
	  if (Stellen > DecimalDigits || Stellen==0) break;
          gewaehlte_Stellenzahl=Stellen;
	  zeile.erase(0, zeile.find(":")+1);
	  rho_Phase=atoi(zeile.substr(0, zeile.find(",")).c_str());
	  zeile.erase(0, zeile.find(",")+1);
	  phi_Phase1=atoi(zeile.substr(0, zeile.find(",")).c_str());
	  zeile.erase(0, zeile.find(",")+1);
	  phi_Phase2=strtod(zeile.substr(0, zeile.find(",")).c_str(),NULL);
	  zeile.erase(0, zeile.find(",")+1);
	  elcu_Phase1=atoi(zeile.substr(0, zeile.find(",")).c_str());
	  zeile.erase(0, zeile.find(",")+1);
	  elcu_Phase2=strtod(zeile.substr(0, zeile.find(",")).c_str(),NULL);
	  zeile.erase(0, zeile.find(",")+1);
	  elcu_Kurven=atoi(zeile.substr(0, zeile.find(",")).c_str());
	  zeile.erase(0, zeile.find(",")+1);
	  StaticFactorbase::Size_StaticFactorbase=atoi(zeile.substr(0, zeile.find(":")).c_str());
	  zeile.erase(0, zeile.find(":")+1);
          Factor_Threshold = atof(zeile.substr(0, zeile.find(":")).c_str());
          if (Factor_Threshold<=0.01 || Factor_Threshold>10.0)
           {
             MARK;
             cerr << "invalid value of Factor_Threshold: " << Factor_Threshold << endl;
             exit(1);
           }
	  zeile.erase(0, zeile.find(":")+1);

	  LogicalSieveSize = atoi(zeile.c_str());
	  LogicalSieveSize = MAX(LogicalSieveSize,PhysicalSieveSize);
	}
    }
  else
    {
       cout << ConfigFile << " not found! Using default values!" << endl;
       static char ch = 'N';
       if (ch!='Y')
         {
           cout << "Bitte mit [Y]es besttigen!" << endl;
           cout << "Please confirm with [Y]es." << endl;
           cin >> ch;
           if (ch!='Y') exit(1);
         }
    }

#ifdef VERBOSE_INFO  
  cout << "Configuration:" << endl;
  cout << "#decimal digits: " << DecimalDigits << endl;
  cout << "actual parameters for " << gewaehlte_Stellenzahl << " digits" << endl;
  cout << "Pollard-Rho: " << rho_Phase << endl;
  cout << "Pollard-Phi: phase 1: " << phi_Phase1 << " phase 2: " << setprecision(1) << phi_Phase2 << endl;
  cout << "elliptic curves: phase 1: " << elcu_Phase1
       << " phase 2: " << elcu_Phase2
       << " #curves: " << elcu_Kurven << endl; 
  cout << "Size of static factorbase: " << StaticFactorbase::Size() << endl;
  cout << "factor-threshold (exponent): " << Factor_Threshold << endl;
  cout << "sieve interval per polynomial: [" << -LogicalSieveSize << "," << LogicalSieveSize << "]" << endl;
#endif

  if (StaticFactorbase::Size()>StaticFactorbase::MaxSize)
    {
      cerr << "MaxSize_StaticFactorbase is too small (or Size_StaticFactorbase is to big)!" << endl;
      exit(1);
    }
}
#endif /* not defined(IS_CLIENT) */
