/*! @file
 * @brief
 * mpqs sieving stuff that needs to be included at compile-time
 */


void do_sieving_Squares()
{
  // Sieve squares of static (and dynamic) factors.
  // Important: Only subtract the single logarithm of the factors
  //            (because if a square divides, then the factor has diveded already!)
  // Remark: To be more efficient, at most one hit in the sieve interval
  //         is considered.

  //cout << "Start sieving squares" << endl;
  if (Sieve_Delta_HeapOfSquares.empty()) return; // there's nothing to sieve...
  TSieve_Delta sieb_delta = Sieve_Delta_HeapOfSquares.top();
  register int delta = sieb_delta.delta-SieveOffset;
  
  while (delta<PhysicalSieveSize)
    {
      Sieve_Delta_HeapOfSquares.pop(); // pop element out of priority queue
      SieveArray[delta]-=static_cast<TSieveElement>(ceil(log_SieveEntryMultiplier*log(sieb_delta.factor))); // mark factor at sieve position
      if (Sieve_Delta_HeapOfSquares.empty()) break;
      sieb_delta = Sieve_Delta_HeapOfSquares.top();
      delta = sieb_delta.delta-SieveOffset;
    }
  //cout << "Finished sieving squares" << endl;
}


#if 0
// old version, obsolete (too many gotos)
void do_sieving_DynamicFactors()
{
  if (Sieve_Delta_Heap.empty()) return; // there's nothing to sieve...
  short int HitCount=0;
  TSieve_Delta sieb_delta = Sieve_Delta_Heap.top();
  register int delta = sieb_delta.delta-SieveOffset;
  goto LoopNext;
  
 LoopEntry_addHit:
  ++HitCount;
#ifdef DEBUG
  if (HitCount>=MaxHits) { cerr << "increase MaxHits!"; exit(1); }
#endif
 LoopEntry_normal:
  Sieve_Delta_Heap.pop(); // pop element out of priority queue
  SieveArray[delta]-=log_DynamicLargePrime(sieb_delta.factor); // mark factor at sieve position
  
  // to speedup refactorization of relations, a dynamic factor hit is memorized here
  Hits[HitCount].Faktor=sieb_delta.factor;

#ifdef USE_FAKEHEAP
  if (Sieve_Delta_Heap.empty()) goto queue_empty; // no further factors to sieve with?
#else
  sieb_delta.delta += sieb_delta.factor; // next hit-position
  if (sieb_delta.delta<=LogicalSieveSize) // shortcut: push only, if it will be inside the sieve interval
   {
     Sieve_Delta_Heap.push(sieb_delta); // pushing to priority queue again
   }
  else
    if (Sieve_Delta_Heap.empty()) goto queue_empty; // no further factors to sieve with?
#endif
  
  sieb_delta = Sieve_Delta_Heap.top(); // get next entry out of the priority queue
  register int delta2; delta2 = sieb_delta.delta-SieveOffset;
  if (delta==delta2) goto LoopEntry_addHit;  
  if (SieveArray[delta]<0) // found a (possible) relation?
   if (SieveControl::Hit_after_DirtyCorrection(delta+SieveOffset,SieveArray[delta]))
    {
      SieveArray[delta]=1; // stall further hits at this position...
      StaticRelations::insert(new CRelation(delta+SieveOffset, HitCount+1)); // and insert as a hit
    }
  delta=delta2;
  
  HitCount=0; // reset HitCount for next hit
 LoopNext:
  if (delta<PhysicalSieveSize) goto LoopEntry_normal; // still in the current (physical) interval?
  return; // done with this interval
  
 queue_empty:
  if (SieveArray[delta]<0) // found a (possible) relation?
   if (SieveControl::Hit_after_DirtyCorrection(delta+SieveOffset,SieveArray[delta]))
    {
      SieveArray[delta]=1; // stall further hits at this position...
      StaticRelations::insert(new CRelation(delta+SieveOffset,HitCount+1)); // and insert as a hit
    }
  return;
}

#else
// new version, patch contributed by Alexis Michon
void do_sieving_DynamicFactors()
{
  if (Sieve_Delta_Heap.empty()) return; // there's nothing to sieve...

  short int HitCount=0;
  TSieve_Delta sieb_delta = Sieve_Delta_Heap.top();
  register int delta = sieb_delta.delta-SieveOffset;

  while (delta < PhysicalSieveSize) // still in the current (physical) interval?
  {
	Sieve_Delta_Heap.pop(); // pop element out of priority queue
	SieveArray[delta]-=log_DynamicLargePrime(sieb_delta.factor); // mark factor at sieve position
  
        // to speedup refactorization of relations, a dynamic factor hit is memorized here
	Hits[HitCount].Faktor=sieb_delta.factor;

#ifdef USE_FAKEHEAP
	if (Sieve_Delta_Heap.empty()) 
	{
		if (SieveArray[delta]<0) // found a (possible) relation?
		if (SieveControl::Hit_after_DirtyCorrection(delta+SieveOffset,SieveArray[delta]))
		{
			SieveArray[delta]=1; // stall further hits at this position...
			StaticRelations::insert(new CRelation(delta+SieveOffset,HitCount+1)); // and insert as a hit
		}
		return;
	}
#else
	sieb_delta.delta += sieb_delta.factor; // next hit-position
	if (sieb_delta.delta<=LogicalSieveSize) // shortcut: push only, if it will be inside the sieve interval
	{
		Sieve_Delta_Heap.push(sieb_delta); // pushing to priority queue again
	}
	else
	if (Sieve_Delta_Heap.empty()) 
	{
		if (SieveArray[delta]<0) // found a (possible) relation?
		 if (SieveControl::Hit_after_DirtyCorrection(delta+SieveOffset,SieveArray[delta]))
		 {
			SieveArray[delta]=1; // stall further hits at this position...
			StaticRelations::insert(new CRelation(delta+SieveOffset,HitCount+1)); // and insert as a hit
		 }
		return;
	}
#endif
  
	sieb_delta = Sieve_Delta_Heap.top(); // get next entry out of the priority queue
	register int delta2; delta2 = sieb_delta.delta-SieveOffset;
	if (delta==delta2)
  	{
		++HitCount;
#ifdef DEBUG
                if (HitCount>=MaxHits) { cerr << "increase MaxHits!"; exit(1); }
#endif
		continue;
	}
	if (SieveArray[delta]<0) // found a (possible) relation?
	 if (SieveControl::Hit_after_DirtyCorrection(delta+SieveOffset,SieveArray[delta]))
	 {
		SieveArray[delta]=1; // stall further hits at this position...
		StaticRelations::insert(new CRelation(delta+SieveOffset, HitCount+1)); // and insert as a hit
	 }
	delta=delta2;
	HitCount=0; // reset HitCount for next hit
  }
  return; // done with this interval
}
#endif


#if 0
#include <map>
void Werteverteilung_im_Sieb_ausgeben(void)
{
  // if you want to view more stats, you can use this function
  // to get the distribution of hits in the sieve
  // (don't activate this for "porduction use", as this is very slow!)
  typedef map<TSieveElement,unsigned int> TStats;
  static TStats S;
  for (int i=0; i<PhysicalSieveSize; ++i) S[SieveArray[i]]++;

  static int z = 1000;
  if (--z==0)
   {
     z=1000;

     for (TStats::iterator pos=S.begin(); pos!=S.end(); ++pos)
      cout << static_cast<signed long int>(pos->first) << " : " << pos->second << endl;

     cout << "press <Enter> to continue" << endl;
     string s; cin >> s;
   }
}
#endif


void recompute_all_Deltas()
{
  // While changing polynomials, the old Deltavalues get void.
  // They must be recomputed.

  //cout << "recomputing deltas" << endl;

  // at first, remove any (possible remaining) dynamic factors from the queue:
#ifdef USE_FAKEHEAP
//  cout << "clearing Fakeheap with " << Sieve_Delta_Heap.size() << " elements..." << endl;
  Sieve_Delta_Heap.clear();
#else
  while (!Sieve_Delta_Heap.empty()) Sieve_Delta_Heap.pop(); // for priority_queue (in case that no clear-method is defined)
#endif

  // do the same with squares:
  //Sieve_Delta_HeapOfSquares.clear();
  while (!Sieve_Delta_HeapOfSquares.empty()) Sieve_Delta_HeapOfSquares.pop(); // for priority_queue (in case that no clear-method is defined)
 
  
  // first, compute the Deltas in/of the static factorbase
  unsigned int inv_A2_modp, PolyB;
  mpz_t inv_A2, P, wuq, x, y;
  mpz_init(inv_A2); mpz_init(P); mpz_init(wuq); mpz_init(x); mpz_init(y);


#if 1 /* Test!!!! 2004-03-09 */
  for (int i=FB_maxQuadrate+1; i<StaticFactorbase::Size() && PrimeNumbers[i]<25000; ++i)
    {
      if (Vorfaktor_von_kN%PrimeNumbers[i]!=0)
	{
          // For all/some "large" squares of the factorbase we determine the Deltas,
          // but these squares are sieved in conjunction with the dynamic factors.
          const unsigned int P = PrimeNumbers[i]*PrimeNumbers[i];
          const unsigned int wuq = SQRT_kN_of_PrimeSquares[i];
          inv_A2_modp=invmod(mpz_remainder_ui(Polynom.get_A2(),P),P);
          PolyB=mpz_remainder_ui(Polynom.get_B(),P);
      
          // Delta1
          signed int d = wuq>=PolyB ? wuq-PolyB : wuq-PolyB+P; // overflow?
          d=mulmod(d,inv_A2_modp,P);
          d+=LogicalSieveSize; d%=P; d-=LogicalSieveSize;
	  if (d<LogicalSieveSize) // value in interval?
	    {
	      TSieve_Delta sieb_delta;
	      sieb_delta.factor=PrimeNumbers[i];
	      sieb_delta.delta=d;
	      Sieve_Delta_HeapOfSquares.push(sieb_delta); // and store
	      //cout << "Square of " << sieb_delta.factor << " pushed!" << endl;
#if defined(DEBUG)
              Polynom.get_values(sieb_delta.delta,x,y);
              if (!mpz_divisible_ui_p(y,P))
               {
                 MARK; cerr << "wrong delta position for square!" << endl; exit(1);
               } //else { cout << "square of " << sieb_delta.factor << " pushed and okay (delta1)" << endl; };
#endif
	    }

	  // Delta2
          unsigned int h = wuq+PolyB; d=P-mulmod(h,inv_A2_modp,P); d+=LogicalSieveSize; d%=P; d-=LogicalSieveSize;
	  if (d<LogicalSieveSize) // value in interval?
	    {
	      TSieve_Delta sieb_delta;
	      sieb_delta.factor=PrimeNumbers[i];
	      sieb_delta.delta=d;
	      Sieve_Delta_HeapOfSquares.push(sieb_delta); // and store
	      //cout << "Square of " << sieb_delta.factor << " pushed!" << endl;
#if defined(DEBUG)
              Polynom.get_values(sieb_delta.delta,x,y);
              if (!mpz_divisible_ui_p(y,P))
               {
                 MARK; cerr << "wrong delta position for square!" << endl; exit(1);
               } //else { cout << "square of " << sieb_delta.factor << " pushed and okay (delta2)" << endl; };
#endif
	    }
	  
	}
    }
#endif



#ifdef SIEVING_LARGE_SQUARES
  for (int i=StaticFactorbase::Size()-1; i>FB_maxQuadrate && PrimeNumbers[i]>25000; --i)
    {
      if (Vorfaktor_von_kN%PrimeNumbers[i]!=0)
	{
          // For all "large" squares of the factorbase we determine the Deltas,
          // but these squares are sieved in conjunction with the dynamic factors.
	  mpz_set_ui(P,PrimeNumbers[i]); mpz_mul(P,P,P); // get square of the prime number
	  mpz_invert(inv_A2,Polynom.get_A2(),P);

          // now compute square roots modulo P (these values are not memorized; this would take too much space)
          // compute square roots of kN modulo Primzahl^2
	  const int Wu = SQRT_kN_of_PrimeNumbers[i];
	  mpz_set_ui(y,2*Wu);
	  if (mpz_invert(y,y,P)==0)
	    {
	      cerr << "PQ-root for " << PrimeNumbers[i] << ": inverse doesn't exist!" << endl;
	      exit(1);
	    }
	  
	  mpz_mod(x,kN,P); mpz_set_ui(wuq,Wu); mpz_mul_ui(wuq,wuq,Wu); mpz_sub(x,x,wuq);
	  mpz_mul(x,x,y); mpz_add_ui(x,x,Wu); mpz_mod(wuq,x,P);

#ifdef DEBUG
          // we may want to check, whether above computations are valid:
          mpz_mul(x,wuq,wuq); mpz_sub(x,kN,x); mpz_mod(x,x,P);
          if (mpz_cmp_ui(x,0)!=0)
           {
             cerr << "PQ-root incorrect!" << endl;
             cerr << "Primzahl=" << PrimeNumbers[i] << endl;
             cerr << "kN= " << kN << endl;
             cerr << "SQRT(kN) mod p=" << SQRT_kN_of_PrimeNumbers[i] << endl;
             cerr << "SQRT(kN) mod p^2=" << wuq << endl;
             cerr << "but we got (wrong value): " << x << endl;
             exit(1); 
           }
#endif
	  // and now the Deltas:

	  // Delta1
	  mpz_sub(x,wuq,Polynom.get_B()); mpz_mul(x,x,inv_A2); mpz_add_ui(x,x,LogicalSieveSize); mpz_mod(x,x,P);
	  mpz_sub_ui(x,x,LogicalSieveSize);
	  if (mpz_cmp_ui(x,LogicalSieveSize)<=0) // does the value lie in the interval?
	    {
	      TSieve_Delta sieb_delta;
	      sieb_delta.factor=PrimeNumbers[i];
	      sieb_delta.delta=mpz_get_si(x);
	      Sieve_Delta_HeapOfSquares.push(sieb_delta); // and store
	      //cout << "Square of " << sieb_delta.factor << " pushed!" << endl;
#ifdef DEBUG
              Polynom.get_values(sieb_delta.delta,x,y);
              if (!mpz_divisible_p(y,P))
               {
                 MARK; cerr << "wrong delta position for square!" << endl; exit(1);
               } //else { cout << "square of " << sieb_delta.factor << " pushed and okay (delta1)" << endl; };
#endif
	    }
	  
	  // Delta2
	  mpz_add(x,wuq,Polynom.get_B()); mpz_neg(x,x); mpz_mul(x,x,inv_A2); mpz_add_ui(x,x,LogicalSieveSize); mpz_mod(x,x,P);
	  mpz_sub_ui(x,x,LogicalSieveSize);
	  if (mpz_cmp_ui(x,LogicalSieveSize)<=0) // value inside interval?
	    {
	      TSieve_Delta sieb_delta;
	      sieb_delta.factor=PrimeNumbers[i];
	      sieb_delta.delta=mpz_get_si(x);
	      Sieve_Delta_HeapOfSquares.push(sieb_delta); // and store
	      //cout << "Square of " << sieb_delta.factor << " pushed!" << endl;
#ifdef DEBUG
              Polynom.get_values(sieb_delta.delta,x,y);
              if (!mpz_divisible_p(y,P))
               {
                 MARK; cerr << "wrong delta position for square!" << endl; exit(1);
               } //else { cout << "square of " << sieb_delta.factor << " pushed and okay (delta2)" << endl; };
#endif
	    }
	  
	}
    }
#endif

  
  // factorbase: compute first sieve hit; ignore primenumber 2, because computation of root for this value is too complicated...
  for (int i=StaticFactorbase::Size()-1; PrimeNumbers[i]>2; --i)
    {
      inv_A2_modp=invmod(mpz_remainder_ui(Polynom.get_A2(),PrimeNumbers[i]),PrimeNumbers[i]);
      PolyB=mpz_remainder_ui(Polynom.get_B(),PrimeNumbers[i]);
      
      register int d;
      d=SQRT_kN_of_PrimeNumbers[i]-PolyB; if (d<0) d+=PrimeNumbers[i];
      Delta_of_PrimeNumbers[i][0]=mulmod(d,inv_A2_modp,PrimeNumbers[i]);
      
      d=PrimeNumbers[i]-SQRT_kN_of_PrimeNumbers[i]-PolyB; if (d<0) d+=PrimeNumbers[i];
      Delta_of_PrimeNumbers[i][1]=mulmod(d,inv_A2_modp,PrimeNumbers[i]);

#ifdef DEBUG
      // we may want to check, whether above computations are valid:
      Polynom.get_values(Delta_of_PrimeNumbers[i][0],x,y);
      if (mpz_mod_ui(y,y,PrimeNumbers[i])!=0) 
	{ 
	  cerr << "Delta0 " << Delta_of_PrimeNumbers[i][0] << " doesn't divide for "
	       << PrimeNumbers[i] << "! (remainder: " << y << ")" << endl;
	  Delta_of_PrimeNumbers[i][0]-=mpz_get_ui(y);
	  if (Delta_of_PrimeNumbers[i][0]<0) Delta_of_PrimeNumbers[i][0]+=PrimeNumbers[i];
	}
      Polynom.get_values(Delta_of_PrimeNumbers[i][1],x,y);
      if (mpz_mod_ui(y,y,PrimeNumbers[i])!=0)
	{ 
	  cerr << "Delta1 " << Delta_of_PrimeNumbers[i][1] << " doesn't divide for "
	       << PrimeNumbers[i] << "! (remainder: " << y << ")" << endl;
	  Delta_of_PrimeNumbers[i][1]-=mpz_get_ui(y);
	  if (Delta_of_PrimeNumbers[i][1]<0) Delta_of_PrimeNumbers[i][1]+=PrimeNumbers[i];
	}
#endif
    }

#ifdef SIEVING_MORE_LARGE_SQUARES
  // and push more squares outside of the static factorbase:
  {
    unsigned int Primzahl = biggest_Prime_in_Factorbase;
    const unsigned int UpperLimit = 1000000; // or maybe 4*biggest_Prime_in_Factorbase;
    cout << "pushing more prime squares up to " << UpperLimit << endl; 
    while (Primzahl<UpperLimit)
     {
       do // compute next prime number of the static factorbase
	{
	  do Primzahl+=2; 
           while( Primzahl%3==0 || Primzahl%5==0 || Primzahl%7==0
                  || Primzahl%11==0 || Primzahl%13==0 || Primzahl%17==0
                  || !probab_prime(Primzahl)); // next prime number
	  mpz_set_ui(P,Primzahl);
	}
       while (mpz_legendre(kN,P)<0); // prime number valid for factorbase?

       mpz_mul(P,P,P); // compute square
       mpz_invert(inv_A2,Polynom.get_A2(),P);

       // now compute square roots modulo P (these values are not memorized; this would take too much space)
       // compute square roots of kN modulo Primzahl^2
       const int Wu = SQRT_kN_mod_PrimeNumber(Primzahl);
       mpz_set_ui(y,2*Wu);
       if (mpz_invert(y,y,P)==0)
        {
          cerr << "PQ-root for " << Primzahl << ": inverse doesn't exist!" << endl;
          exit(1);
	}
	  
       mpz_mod(x,kN,P); mpz_set_ui(wuq,Wu); mpz_mul_ui(wuq,wuq,Wu); mpz_sub(x,x,wuq);
       mpz_mul(x,x,y); mpz_add_ui(x,x,Wu); mpz_mod(wuq,x,P);

#ifdef DEBUG
       // we may want to check, whether above computations are valid:
       mpz_mul(x,wuq,wuq); mpz_sub(x,kN,x); mpz_mod(x,x,P);
       if (mpz_cmp_ui(x,0)!=0)
        {
          cerr << "PQ-root incorrect!" << endl;
          cerr << "Primzahl=" << Primzahl << endl;
          cerr << "kN= " << kN << endl;
          cerr << "SQRT(kN) mod p=" << Wu << endl;
          cerr << "SQRT(kN) mod p^2=" << wuq << endl;
          cerr << "but we got (wrong value): " << x << endl;
          exit(1); 
        }
#endif
       // and now compute the Deltas:

       // Delta1
       mpz_sub(x,wuq,Polynom.get_B()); mpz_mul(x,x,inv_A2); mpz_add_ui(x,x,LogicalSieveSize); mpz_mod(x,x,P);
       mpz_sub_ui(x,x,LogicalSieveSize);
       if (mpz_cmp_ui(x,LogicalSieveSize)<=0) // value inside interval?
        {
          TSieve_Delta sieb_delta;
          sieb_delta.factor=Primzahl;
          sieb_delta.delta=mpz_get_si(x);
          Sieve_Delta_HeapOfSquares.push(sieb_delta); // and store

          // IMPORTANT: The prime number will be placed into Sieve_Delta_HeapOfSquares *and*
          // additionally into Sieve_Delta_Heap (if it is a square of a unknown/unresolved dynamic factor).
          // Please remember: Known dynamic factors will be sieved anyway, so sieving two times again
          // for squares should be avoided.
          // Doing this, we can assure, that
          //  - the value (factor instead of factor^2) fits into integer range,
          //  - the value is hitting twice (as square) inside the sieve.
          // The (possible) effect, that a square is counted with 3 hits remains still possible (whenever
          // a new dynamic factor is detected during sieving and deactivated FAKEHEAP), but this case
          // is unlikely and harmless. (In the worst case a non-relation will falsely be considered as
          // a relation and rejected afterwards.)
          if (!is_dynamic_factor(Primzahl)) Sieve_Delta_Heap.push(sieb_delta); // and store
#ifdef VERBOSE
          cout << "Pushed square of " << sieb_delta.factor << " (Delta1)!" << endl;
#endif
#ifdef DEBUG
          Polynom.get_values(sieb_delta.delta,x,y);
          if (!mpz_divisible_p(y,P))
           {
             MARK; cerr << "wrong delta position for square!" << endl; exit(1);
           } else { cout << "large square of " << Primzahl << " pushed and okay (delta1)" << endl; };
#endif
        }

       // Delta2
       mpz_add(x,wuq,Polynom.get_B()); mpz_neg(x,x); mpz_mul(x,x,inv_A2); mpz_add_ui(x,x,LogicalSieveSize); mpz_mod(x,x,P);
       mpz_sub_ui(x,x,LogicalSieveSize);
       if (mpz_cmp_ui(x,LogicalSieveSize)<=0) // value inside interval?
        {
          TSieve_Delta sieb_delta;
          sieb_delta.factor=Primzahl;
          sieb_delta.delta=mpz_get_si(x);
          Sieve_Delta_HeapOfSquares.push(sieb_delta); // and push into priority queue
          // see remark for Delta1
          if (!is_dynamic_factor(Primzahl)) Sieve_Delta_Heap.push(sieb_delta); // push into priority queue
#ifdef VERBOSE
          cout << "Pushed square of " << sieb_delta.factor << " (Delta2)!" << endl;
#endif
#ifdef DEBUG
          Polynom.get_values(sieb_delta.delta,x,y);
          if (!mpz_divisible_p(y,P))
           {
             MARK; cerr << "wrong delta position for square!" << endl; exit(1);
           } else { cout << "large square of " << Primzahl << " pushed and okay (delta2)" << endl; };
#endif
        }

     }
  }
#endif

#if defined(VERBOSE) || defined(DEBUG)
  if (!Sieve_Delta_HeapOfSquares.empty())
   cout << Sieve_Delta_HeapOfSquares.size() << " squares have been pushed." << endl;
#endif
  
  for (int i=0; i<NumberOf_more_PrimePowers; ++i)
    {
      // compute Deltas for these numbers (which are powers of prime numbers of the factorbase)
      // (and sieve them together with the static factors)

      inv_A2_modp=invmod(mpz_remainder_ui(Polynom.get_A2(),PrimePowers[i]),PrimePowers[i]);
      PolyB=mpz_remainder_ui(Polynom.get_B(),PrimePowers[i]);

      register int d;
      d=SQRT_kN_of_PrimePowers[i]-PolyB; if (d<0) d+=PrimePowers[i];
      Delta_of_PrimePowers[i][0]=mulmod(d,inv_A2_modp,PrimePowers[i]);
      
      d=PrimePowers[i]-SQRT_kN_of_PrimePowers[i]-PolyB; if (d<0) d+=PrimePowers[i];
      Delta_of_PrimePowers[i][1]=mulmod(d,inv_A2_modp,PrimePowers[i]);
    }
  
  mpz_clear(inv_A2); mpz_clear(P); mpz_clear(wuq); mpz_clear(x); mpz_clear(y);
  
  // and now do the same computations for dynamic factorbase:
  TDynamicFactorRelations::iterator pos;
  for (pos=DynamicFactorRelations.begin();
       pos!=DynamicFactorRelations.end() && pos->sieveable(); ++pos) compute_Deltas_for(*pos, -LogicalSieveSize);

#if 0
  // we are a little bit disobedient,
  // we're sieving more than "sieveable()" allows,
  // up to StaticFactorbase::Size()/4 more...
  int more;
  for (more=StaticFactorbase::Size()>>2;
       pos!=DynamicFactorRelations.end() && more; ++pos,--more)
     compute_Deltas_for(*pos, -LogicalSieveSize);
  //cout << "disobedient: " << (static_cast<int>(StaticFactorbase::Size()>>2))-more << endl;
#endif
  
 // cout << "Finished computation of Deltas." << endl;
}
